package de.fraunhofer.iem.dataprovider.taskManager

import de.fraunhofer.iem.dataprovider.logger.getLogger
import kotlinx.coroutines.*
import kotlinx.coroutines.channels.Channel
import org.springframework.stereotype.Component

enum class TaskType{
    REPO_CHANGED,
    DONE
}
data class Task(val type: TaskType, val payload: Any)
@Component
class TaskManager : Worker("Task Manager") {

    private val gitWorker = GitWorker(tasks)

    override suspend fun executeTask(task: Task) {
        when (task.type) {
            TaskType.REPO_CHANGED -> {
                logger.info("[${Thread.currentThread().name}] add task called in ${task.payload}")
                gitWorker.addTask(task)
                logger.info("[${Thread.currentThread().name}] add task finished in ${task.payload}")
            }
            TaskType.DONE -> {
                logger.info("${Thread.currentThread().name}] Task done ${task.payload}")
            }
        }
    }
}

class GitWorker(private val responseChannel: Channel<Task>): Worker("Git Worker") {

    override suspend fun executeTask(task: Task) {
        // simulate a long running task
        logger.info("[${Thread.currentThread().name}] Starting task ${task.payload} in gitworker")
        delay(5000)
        responseChannel.send(Task(TaskType.DONE, "Task with payload ${task.payload} finished"))
        logger.info("[${Thread.currentThread().name}] Finished task ${task.payload} in gitworker")
    }

}
abstract class Worker(private val name: String): IWorker {

    // TODO: this could be dangerous to give it unlimited memory
    protected val tasks = Channel<Task>(Channel.UNLIMITED)
    private val numCores = Runtime.getRuntime().availableProcessors()
    // use a fixed size thread pool context with numCores threads
    private val threadPoolContext = newFixedThreadPoolContext(numCores, name)
    private val coroutineScope = CoroutineScope(threadPoolContext)

    protected val logger = getLogger(javaClass)
    override suspend fun addTask(task: Task) {
        tasks.send(task)
    }

    init {
        logger.info("Starting $numCores workers.")
        repeat(numCores) {
            logger.info("[${Thread.currentThread().name}] launching coroutine in $name")
            launchWorker(it)
        }
    }

    private fun launchWorker(id: Int) = coroutineScope.launch {

        for (task in tasks) {
            coroutineScope.launch {
                logger.debug("[${Thread.currentThread().name}] Processor #$id-$name received $task")
                executeTask(task)
                logger.debug("[${Thread.currentThread().name}] Processor #$id-$name finished $task")
            }
        }
    }

    override fun close() {
        coroutineScope.launch {
            tasks.close()
            for (job in coroutineContext[Job]!!.children) {
                job.join()
            }
            threadPoolContext.close()
            logger.info("Worker $name has been closed")
        }
    }
}

interface IWorker {
    suspend fun addTask(task: Task)
    suspend fun executeTask(task: Task)
    fun close()
}